@@ -2,32 +2,136 @@
 
 """Field definitions for the flexible_forms module."""
 
-from typing import Any, Dict, Mapping, Optional, Sequence, Tuple, Type
+import logging
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Dict,
+    List,
+    Mapping,
+    Optional,
+    Sequence,
+    Tuple,
+    Type,
+    cast,
+)
 
+import jmespath
+import requests
 import simpleeval
+from django.core.paginator import Paginator
 from django.db.models import FileField, ImageField
 from django.db.models import fields as model_fields
 from django.forms import fields as form_fields
 from django.forms import widgets as form_widgets
+from django.http.request import HttpRequest
+from django.shortcuts import get_object_or_404
+from django.template import Context, Template
+from django.template.base import VariableNode
+from django.urls import reverse
+
+from flexible_forms.widgets import (
+    AutocompleteResult,
+    AutocompleteSelect,
+    AutocompleteSelectMultiple,
+)
 
 try:
     from django.db.models import JSONField  # type: ignore
+    from django.forms import JSONField as JSONFormField  # type: ignore
 except ImportError:  # pragma: no cover
     from django.contrib.postgres.fields import JSONField
+    from django.contrib.postgres.forms import JSONField as JSONFormField
 
-from flexible_forms.utils import all_subclasses, evaluate_expression
+from flexible_forms.utils import evaluate_expression, stable_json
+
+logger = logging.getLogger(__name__)
 
 ##
 # _EMPTY_CHOICES
 _EMPTY_CHOICES = (("EMPTY", "Select a value."),)
 
+if TYPE_CHECKING:  # pragma: no cover
+    from flexible_forms.models import BaseField, BaseRecord
 
-class FlexibleField:
+
+##
+# FIELD_TYPES
+#
+# A dict mapping of field types, where the key is the string representation of
+# the field type (usually its class name), and the value is the `FlexibleField`
+# class itself.
+#
+# Built dynamically to include all descendants of `FieldType`.
+#
+FIELD_TYPES: Dict[str, Type["FieldType"]] = {}
+
+
+class FieldTypeOptions:
+    """A base class for the Meta inner class for FieldType."""
+
+    abstract = False
+    force_replacement = False
+
+
+class FieldTypeMetaclass(type):
+    """A metaclass for handling FieldType configuration and registration."""
+
+    def __new__(
+        cls,
+        name: str,
+        bases: Tuple[type, ...],
+        attrs: Dict[str, Any],
+        **kwargs: Any,
+    ) -> "FieldTypeMetaclass":
+        """Create a new FieldType and register it.
+
+        Args:
+            cls: This metaclass.
+            name: The name of the new FieldType class.
+            bases: The base classes of the new FieldType class.
+            attrs: The attributes of the new FieldType class.
+            kwargs: Passed to super.
+
+        Returns:
+            Type[FieldType]: The new FieldType class, configured and registered.
+
+        Raises:
+            ValueError: If a FieldType with the same class name was already
+                registered, and the force_replacement meta option is False.
+        """
+        attrs_meta = attrs.pop("Meta", None)
+        attrs["_meta"] = type(
+            "Meta", tuple(filter(bool, (attrs_meta, FieldTypeOptions))), {}
+        )
+
+        # Set the name from the class name if a name was not provided.
+        attrs["name"] = attrs.get("name", name)
+
+        clsobj = super().__new__(cls, name, bases, attrs, **kwargs)  # type: ignore
+
+        # Throw an error if a FieldType with the given name was already registered.
+        if clsobj.name in FIELD_TYPES and not clsobj._meta.force_replacement:
+            raise ValueError(
+                f"A FieldType named {name} was already registered ({FIELD_TYPES[name]})."
+            )
+
+        # Register the new field type with its class name (if it's not abstract).
+        if not clsobj._meta.abstract:
+            FIELD_TYPES[name] = clsobj
+
+        return cast("FieldTypeMetaclass", clsobj)
+
+
+class FieldType(metaclass=FieldTypeMetaclass):
     """A prefabricated field to use with flexible forms.
 
     Provides an interface to emit a model field, form field, or widget.
     """
 
+    class Meta:
+        abstract = True
+
     ##
     # label
     #
@@ -91,55 +195,38 @@ class FlexibleField:
     #
     model_field_options: Dict[str, Any] = {}
 
-    @classmethod
-    def name(cls) -> str:
-        """Return the string representation of the field type.
-
-        Useful for creating a type lookup, or for storing in the database.
-
-        Returns:
-            str: The string representation of the field type.
-        """
-        return cls.__name__
-
     @classmethod
     def as_form_field(
         cls,
+        *,
+        field: "BaseField",
+        record: Optional["BaseRecord"] = None,
         modifiers: Sequence[Tuple[str, str]] = (),
         field_values: Optional[Mapping[str, Optional[Any]]] = None,
-        form_widget_options: Optional[Mapping[str, Optional[Any]]] = None,
         **kwargs: Any,
     ) -> form_fields.Field:
         """Return an instance of the field for use in a Django form.
 
-        Receives a dict of kwargs to pass through to the form field constructor.
+        Passes additional kwargs to the form field constructor.
 
         Args:
-            modifiers: A sequence of modifiers to
-                be applied to the field.
-            field_values: A mapping of the
-                current form values.
-            form_widget_options: A mapping of
-                options to pass to the widget constructor if a widget is
-                configured.
-            kwargs: A dict of kwargs to be passed to the constructor of the
-                `form_field_class`.
+            field: The Field model instance.
+            record: The record instance to which the form is bound, if
+                available.
+            modifiers: A sequence of modifiers to be applied to the field.
+            field_values: A mapping of the current form values.
+            kwargs: Passed to the form field constructor.
 
         Returns:
             form_fields.Field: An instance of the form field.
         """
-        # Generate an appropriate form widget for the field.
-        FormWidget = cls.form_widget_class or cls.form_field_class.widget
-        form_widget = FormWidget(
-            **{
-                **cls.form_widget_options,
-                **(form_widget_options or {}),
-            }
+        # If no widget was given explicitly, build a default one.
+        kwargs["widget"] = kwargs.get(
+            "widget", cls.as_form_widget(field=field, record=record)
         )
 
         # Generate the form field with its appropriate class and widget.
         form_field = cls.form_field_class(
-            widget=form_widget,
             **{
                 **cls.form_field_options,
                 **kwargs,
@@ -155,6 +242,32 @@ class FlexibleField:
 
         return form_field
 
+    @classmethod
+    def as_form_widget(
+        cls, field: "BaseField", record: Optional["BaseRecord"] = None, **kwargs: Any
+    ) -> form_widgets.Widget:
+        """Return an instance of the form widget for rendering.
+
+        Passes additional kwargs to the form widget constructor.
+
+        Args:
+            field: The Field model instance.
+            record: The record instance to which the form is bound, if
+                available.
+            kwargs: Passed through to the form widget constructor.
+
+        Returns:
+            form_widgets.Widget: The configured form widget for the field.
+        """
+        widget_cls = cls.form_widget_class or cls.form_field_class.widget
+
+        return widget_cls(
+            **{
+                **cls.form_widget_options,
+                **kwargs,
+            }
+        )
+
     @classmethod
     def as_model_field(cls, **kwargs: Any) -> model_fields.Field:
         """Return an instance of the field for use in a Django model.
@@ -260,7 +373,7 @@ class FlexibleField:
         return form_field
 
 
-class SingleLineTextField(FlexibleField):
+class SingleLineTextField(FieldType):
     """A field for collecting single-line text values."""
 
     label = "Single-line Text"
@@ -269,7 +382,7 @@ class SingleLineTextField(FlexibleField):
     model_field_class = model_fields.TextField
 
 
-class MultiLineTextField(FlexibleField):
+class MultiLineTextField(FieldType):
     """A field for collecting multiline text values."""
 
     label = "Multi-line Text"
@@ -279,7 +392,7 @@ class MultiLineTextField(FlexibleField):
     model_field_class = model_fields.TextField
 
 
-class EmailField(FlexibleField):
+class EmailField(FieldType):
     """A field for collecting email addresses."""
 
     label = "Email Address"
@@ -288,7 +401,7 @@ class EmailField(FlexibleField):
     model_field_class = model_fields.EmailField
 
 
-class URLField(FlexibleField):
+class URLField(FieldType):
     """A field for collecting URL values."""
 
     label = "URL"
@@ -298,7 +411,7 @@ class URLField(FlexibleField):
     model_field_options = {"max_length": 2083}
 
 
-class SensitiveTextField(FlexibleField):
+class SensitiveTextField(FieldType):
     """A field for collecting sensitive text values.
 
     Primarily useful for obfuscating the input when on-screen.
@@ -311,7 +424,7 @@ class SensitiveTextField(FlexibleField):
     model_field_class = model_fields.TextField
 
 
-class IntegerField(FlexibleField):
+class IntegerField(FieldType):
     """A field for collecting integer values."""
 
     label = "Integer"
@@ -321,7 +434,7 @@ class IntegerField(FlexibleField):
     model_field_class = model_fields.IntegerField
 
 
-class PositiveIntegerField(FlexibleField):
+class PositiveIntegerField(FieldType):
     """A field for collecting positive integer values."""
 
     label = "Positive Integer"
@@ -331,7 +444,7 @@ class PositiveIntegerField(FlexibleField):
     model_field_class = model_fields.PositiveIntegerField
 
 
-class DecimalField(FlexibleField):
+class DecimalField(FieldType):
     """A field for collecting decimal number values."""
 
     label = "Decimal Number"
@@ -342,7 +455,7 @@ class DecimalField(FlexibleField):
     model_field_options = {"max_digits": 15, "decimal_places": 6}
 
 
-class DateField(FlexibleField):
+class DateField(FieldType):
     """A field for collecting date data."""
 
     label = "Date"
@@ -351,7 +464,7 @@ class DateField(FlexibleField):
     model_field_class = model_fields.DateField
 
 
-class TimeField(FlexibleField):
+class TimeField(FieldType):
     """A field for collecting time data."""
 
     label = "Time"
@@ -360,7 +473,7 @@ class TimeField(FlexibleField):
     model_field_class = model_fields.TimeField
 
 
-class DateTimeField(FlexibleField):
+class DateTimeField(FieldType):
     """A field for collecting datetime data."""
 
     label = "Date & Time"
@@ -369,7 +482,7 @@ class DateTimeField(FlexibleField):
     model_field_class = model_fields.DateTimeField
 
 
-class DurationField(FlexibleField):
+class DurationField(FieldType):
     """A field for collecting duration data."""
 
     label = "Duration"
@@ -378,7 +491,7 @@ class DurationField(FlexibleField):
     model_field_class = model_fields.DurationField
 
 
-class CheckboxField(FlexibleField):
+class CheckboxField(FieldType):
     """A field for collecting a boolean value with a checkbox."""
 
     label = "Checkbox"
@@ -388,7 +501,7 @@ class CheckboxField(FlexibleField):
     model_field_class = model_fields.BooleanField
 
 
-class YesNoRadioField(FlexibleField):
+class YesNoRadioField(FieldType):
     """A field for collecting a yes/no value with radio buttons."""
 
     label = "Yes/No Radio Buttons"
@@ -405,7 +518,7 @@ class YesNoRadioField(FlexibleField):
     model_field_class = model_fields.BooleanField
 
 
-class YesNoUnknownRadioField(FlexibleField):
+class YesNoUnknownRadioField(FieldType):
     """A field for collecting a yes/no/unknown value with radio buttons."""
 
     label = "Yes/No/Unknown Radio Buttons"
@@ -426,7 +539,7 @@ class YesNoUnknownRadioField(FlexibleField):
     model_field_options = {"null": True}
 
 
-class YesNoSelectField(FlexibleField):
+class YesNoSelectField(FieldType):
     """A field for collecting a boolean value with a yes/no select field."""
 
     label = "Yes/No Dropdown"
@@ -443,7 +556,7 @@ class YesNoSelectField(FlexibleField):
     model_field_class = model_fields.BooleanField
 
 
-class YesNoUnknownSelectField(FlexibleField):
+class YesNoUnknownSelectField(FieldType):
     """A field for collecting a boolean value with a yes/no select field."""
 
     label = "Yes/No/Unknown Dropdown"
@@ -463,7 +576,7 @@ class YesNoUnknownSelectField(FlexibleField):
     model_field_options = {"null": True}
 
 
-class SingleChoiceSelectField(FlexibleField):
+class SingleChoiceSelectField(FieldType):
     """A field for collecting a single text value from a select list."""
 
     label = "Single-choice Dropdown"
@@ -473,7 +586,7 @@ class SingleChoiceSelectField(FlexibleField):
     model_field_class = model_fields.TextField
 
 
-class SingleChoiceRadioSelectField(FlexibleField):
+class SingleChoiceRadioSelectField(FieldType):
     """A field for collecting a text value from a set of radio buttons."""
 
     label = "Radio Buttons"
@@ -484,7 +597,7 @@ class SingleChoiceRadioSelectField(FlexibleField):
     model_field_class = model_fields.TextField
 
 
-class MultipleChoiceSelectField(FlexibleField):
+class MultipleChoiceSelectField(FieldType):
     """A field for collecting multiple text values from a select list."""
 
     label = "Multiple-choice Dropdown"
@@ -494,7 +607,7 @@ class MultipleChoiceSelectField(FlexibleField):
     model_field_class = JSONField
 
 
-class MultipleChoiceCheckboxField(FlexibleField):
+class MultipleChoiceCheckboxField(FieldType):
     """A field for collecting multiple text values from a checkbox list."""
 
     label = "Multiple-choice Checkboxes"
@@ -505,7 +618,7 @@ class MultipleChoiceCheckboxField(FlexibleField):
     model_field_class = JSONField
 
 
-class FileUploadField(FlexibleField):
+class FileUploadField(FieldType):
     """A field for collecting file uploads."""
 
     label = "File Upload"
@@ -514,7 +627,7 @@ class FileUploadField(FlexibleField):
     model_field_class = FileField
 
 
-class ImageUploadField(FlexibleField):
+class ImageUploadField(FieldType):
     """A field for collecting image uploads."""
 
     label = "Image Upload"
@@ -523,15 +636,184 @@ class ImageUploadField(FlexibleField):
     model_field_class = ImageField
 
 
-##
-# FIELD_TYPES
-#
-# A dict mapping of field types, where the key is the string representation of
-# the field type (usually its class name), and the value is the `FlexibleField`
-# class itself.
-#
-# Built dynamically to include all descendants of `FlexibleField`.
-#
-FIELD_TYPES: Dict[str, Type[FlexibleField]] = {
-    f.name(): f for f in all_subclasses(FlexibleField)
-}
+class AutocompleteSelectField(FieldType):
+    """A field for autocompleting selections from an HTTP endpoint."""
+
+    label = "Autocomplete"
+
+    form_field_class = JSONFormField
+    form_widget_class = AutocompleteSelect
+    model_field_class = JSONField
+
+    @classmethod
+    def as_form_widget(
+        cls,
+        field: "BaseField",
+        record: Optional["BaseRecord"] = None,
+        **form_widget_options: Any,
+    ) -> form_widgets.Widget:
+        """Build the autocomplete form widget.
+
+        Replaces the url parameter with a proxy URL to a custom view to allow
+        developers to implement fine-grained search behavior.
+
+        Args:
+            field: The Field record.
+            record: The record instance to which the form is bound, if
+                available.
+            form_widget_options: Parameters passed to the form widget
+                constructor.
+
+        Returns:
+            form_widgets.Widget: A configured widget for rendering the
+                autocomplete input.
+        """
+        proxy_url = reverse(
+            "flexible_forms:autocomplete",
+            kwargs={
+                "app_label": field._meta.app_label,
+                "model_name": field._meta.model_name,
+                "field_pk": field.pk,
+            },
+        )
+
+        # If a specific record instance was supplied, append its ID to the URL
+        # so that autocomplete suggestions can be tailored to a particular
+        # record's field values.
+        proxy_url = f"{proxy_url}?record_pk={record.pk}" if record else proxy_url
+
+        return super().as_form_widget(
+            field=field,
+            record=record,
+            **{
+                **form_widget_options,
+                "url": proxy_url,
+            },
+        )
+
+    @classmethod
+    def autocomplete(
+        cls,
+        request: HttpRequest,
+        field: "BaseField",
+        **form_widget_options: Any,
+    ) -> Tuple[List[AutocompleteResult], bool]:
+        """Perform a search against the configured URL.
+
+        Returns a sequence of dicts, each with "id" and "text" values to use
+        within a select form element.
+
+        Args:
+            request: The HTTP request.
+            field: The Field record instance that uses this autocomplete.
+            form_widget_options: A dict of form widget options that can be
+                used to customize autocomplete behavior.
+
+        Returns:
+            Tuple[Sequence[AutocompleteResult], bool]: A two-tuple containing
+                a sequence of zero or more select2-compatible search results, and
+                a boolean indicating whether or not there are more results to
+                fetch.
+        """
+        # Extract pagination info.
+        page = int(request.GET.get("page", "1"))
+        per_page = int(request.GET.get("per_page", "100"))
+
+        url = form_widget_options.get("url") or ""
+        results_path = form_widget_options.get("results_path") or "@"
+        text_path = form_widget_options.get("text_path") or "(text || label || name)"
+        id_path = form_widget_options.get("id_path") or "id"
+
+        # Initialize results with an empty list.
+        results: List[AutocompleteResult] = []
+        has_more = False
+
+        # If no URL has been configured, return results as-is.
+        if not url:
+            return results, has_more
+
+        # Prevent a circular import.
+        from flexible_forms.models import BaseRecord
+
+        # If autocompletion was requested for a specific record, fetch it using
+        # the given primary key. Otherwise, make a blank record.
+        record_pk = request.GET.get("record_pk", None)
+        record_model = field.flexible_forms.get_model(BaseRecord)
+        record = (
+            get_object_or_404(record_model, pk=record_pk)
+            if record_pk is not None
+            else record_model(form=field.form)
+        )
+
+        # Render the URL as a Django Template, feeding it the current field
+        # values, the record instance ("meta"), and the GET parameters.
+        url_template = Template(url)
+        rendered_url = url_template.render(
+            Context({"record": record, **request.GET.dict()})
+        )
+
+        # Get a set of all variable names used in the template so that we can
+        # use them for optimizations later.
+        url_vars = frozenset(
+            v.filter_expression.var.lookups[0]
+            for v in url_template.nodelist
+            if isinstance(v, VariableNode)
+        )
+
+        # Search the endpoint and raise an exception if we get any non-success
+        # status in the response.
+        response = requests.get(rendered_url)
+        response.raise_for_status()
+
+        # Decode the JSON response.
+        response_json = response.json()
+
+        # Parse the search response and map each result to a Select2-compatible
+        # dict with an "id" and a "text" property.
+        results_expr = jmespath.compile(results_path)
+        text_expr = jmespath.compile(text_path)
+        id_expr = jmespath.compile(id_path)
+
+        raw_results = results_expr.search(response_json)
+        for result in raw_results:
+            result_text = text_expr.search(result)
+            result_value = id_expr.search(result)
+
+            if result_value is None:
+                result_value = result_text
+
+            # The "id" is the value eventually stored in the database. In this
+            # case, it is a stringified version of the result JSON so that it can
+            # be deserialized when rendering the initial value for the widget.
+            result_id = stable_json(
+                {"id": str(result_value), "text": str(result_text), "extra": {}}
+            )
+
+            # The use of "id" as opposed to "value" or something that makes
+            # more semantic sense is to support the use of the select2
+            # implementation that ships with the Django admin.
+            results.append({"id": result_id, "text": result_text})
+
+        # If the remote endpoint isn't handling pagination, do it manually.
+        if "page" not in url_vars:
+            paginated_page = Paginator(results, per_page).page(page)
+            results, has_more = (
+                cast(List[AutocompleteResult], paginated_page.object_list),
+                paginated_page.has_next(),
+            )
+
+        # If the remote endpoint is handling pagination for us, we can assume
+        # that there are more results if the number of results received is
+        # greater than or equal to our per_page value (i.e., we received a full
+        # page of results).
+        else:
+            has_more = len(results) >= per_page
+
+        return results, has_more
+
+
+class AutocompleteSelectMultipleField(AutocompleteSelectField):
+    """An autocomplete field that supports multiple selections."""
+
+    form_field_class = JSONFormField
+    form_widget_class = AutocompleteSelectMultiple
