@@ -24,7 +24,11 @@ from typing import (
 )
 
 from django import forms
-from django.core.exceptions import ImproperlyConfigured, ValidationError
+from django.core.exceptions import (
+    ImproperlyConfigured,
+    ObjectDoesNotExist,
+    ValidationError,
+)
 from django.core.serializers.json import DjangoJSONEncoder
 from django.db import models, transaction
 from django.db.models.fields.mixins import FieldCacheMixin
@@ -42,7 +46,7 @@ from django.utils.functional import cached_property
 from django.utils.text import slugify
 from simpleeval import FunctionNotDefined, NameNotDefined
 
-from flexible_forms.fields import FIELD_TYPES
+from flexible_forms.fields import FIELD_TYPES, FieldType
 from flexible_forms.utils import (
     FormEvaluator,
     evaluate_expression,
@@ -59,8 +63,8 @@ except ImportError:  # pragma: no cover
 if TYPE_CHECKING:  # pragma: no cover
     from flexible_forms.forms import BaseRecordForm
 
-T = TypeVar(
-    "T",
+M = TypeVar(
+    "M",
     bound="FlexibleBaseModel",
 )
 
@@ -275,6 +279,16 @@ class FlexibleRelation(ForeignObject):
 
         super().__init__(to, *args, **kwargs)
 
+    @property
+    def target_field(self) -> models.Field:
+        """Return the target field for the relation."""
+        try:
+            return super().target_field
+        except IndexError:  # pragma: no cover
+            if TYPE_CHECKING:
+                return cast("models.Field", None)
+            raise  # type: ignore
+
     def contribute_to_class(
         self, cls: Type[models.Model], name: str, private_only: bool = False
     ) -> None:
@@ -400,7 +414,7 @@ class FlexibleBaseModel(models.Model):
     flexible_forms: "FlexibleForms"
 
     @classmethod
-    def _flexible_model_for(cls, base_model: Type[T]) -> Type[T]:
+    def _flexible_model_for(cls, base_model: Type[M]) -> Type[M]:
         """Return the current implementation of the given base_model.
 
         Args:
@@ -410,7 +424,7 @@ class FlexibleBaseModel(models.Model):
             Type[FleixbleModel]: The model class for the given base model in
                 the current implementation.
         """
-        return cast(Type[T], cls.flexible_forms.get_model(base_model))
+        return cls.flexible_forms.get_model(base_model)
 
 
 @receiver(class_prepared)
@@ -478,7 +492,7 @@ class BaseForm(FlexibleBaseModel):
                     f.name: f.initial if isinstance(f.initial, list) else [f.initial]
                     for f in self.fields.all()
                 },
-                "_form": [self],
+                "form": [self],
             }
         )
 
@@ -580,12 +594,7 @@ class BaseForm(FlexibleBaseModel):
             f.name: f.as_form_field() for f in all_fields
         }
         for field_name, form_field in form_fields.items():
-            field_widget = cast(Widget, form_field.widget)
-            field_value = field_widget.value_from_datadict(
-                data=form_state,  # type: ignore
-                files=form_state,
-                name=field_name,
-            )
+            field_value = form_state.get(field_name)
             # Try to perform as much of the value coercion process as possible
             # while attempting to avoid running expensive validators.
             try:
@@ -598,10 +607,10 @@ class BaseForm(FlexibleBaseModel):
 
         # Regenerate the form fields, this time taking the field values into
         # account in order to inform any dynamic behaviors.
-        form_fields = {}
-        for field in all_fields:
-            form_field = field.as_form_field(field_values=form_state)
-            form_fields[field.name] = form_field
+        form_fields = {
+            f.name: f.as_form_field(field_values=form_state, record=instance)
+            for f in all_fields
+        }
 
         RecordModel = self._flexible_model_for(BaseRecord)
 
@@ -623,7 +632,7 @@ class BaseForm(FlexibleBaseModel):
             },
         )
 
-        initial = {"_form": self, **(initial or {})}
+        initial = {"form": self, **(initial or {})}
 
         # Create a form instance from the form class and the passed parameters.
         form_instance = form_class(
@@ -736,36 +745,59 @@ class BaseField(FlexibleBaseModel):
 
         super().save(*args, **kwargs)
 
+    def as_field_type(self) -> Type[FieldType]:
+        """Return the flexible FieldType for the field.
+
+        Returns:
+            FieldType: The FieldType class for the field.
+        """
+        return FIELD_TYPES[self.field_type]
+
     def as_form_field(
         self,
         field_values: Optional[Mapping[str, Any]] = None,
+        record: Optional["BaseRecord"] = None,
     ) -> forms.Field:
-        """Return a Django form Field definition.
+        """Return a Django form Field instance.
 
         Args:
-            field_values: The current values of
-                all fields in the form.
+            field_values: The current values of all fields in the form.
+            record: The record instance that the form is bound to, if
+                available.
 
         Returns:
             forms.Field: The configured Django form Field instance.
         """
-        return FIELD_TYPES[self.field_type].as_form_field(
-            **{
-                # Special parameters.
-                "modifiers": (
-                    (m.attribute, m.expression)
-                    for m in self.modifiers.all()  # type: ignore
-                ),
-                "field_values": field_values,
-                "form_widget_options": self.form_widget_options,
-                # Django form field arguments.
-                "required": self.required,
-                "label": self.label,
-                "label_suffix": self.label_suffix,
-                "initial": self.initial,
-                "help_text": self.help_text,
-                **self.form_field_options,
-            }
+        return self.as_field_type().as_form_field(
+            field=self,
+            record=record,
+            modifiers=tuple((m.attribute, m.expression) for m in self.modifiers.all()),
+            field_values=field_values,
+            # Django form field arguments.
+            required=self.required,
+            label=self.label,
+            label_suffix=self.label_suffix,
+            initial=self.initial,
+            help_text=self.help_text,
+            widget=self.as_form_widget(record=record),
+            **self.form_field_options,
+        )
+
+    def as_form_widget(self, record: Optional["BaseRecord"] = None) -> Widget:
+        """Return a Django form Widget instance.
+
+        Uses the form_widget_options and the given field_values to customize
+        the widget.
+
+        Args:
+            record: The record instance that the form is bound to, if
+                available.
+
+        Returns:
+            Widget: The configured Django form Widget instance.
+        """
+        return self.as_field_type().as_form_widget(
+            field=self, record=record, **self.form_widget_options
         )
 
     def as_model_field(self) -> models.Field:
@@ -774,7 +806,7 @@ class BaseField(FlexibleBaseModel):
         Returns:
             models.Field: The configured Django model Field instance.
         """
-        return FIELD_TYPES[self.field_type].as_model_field(
+        return self.as_field_type().as_model_field(
             null=not self.required,
             blank=not self.required,
             default=self.initial,
@@ -843,23 +875,38 @@ class BaseFieldModifier(FlexibleBaseModel):
         super().clean()
         self._validated = False
 
-        # No field? No Problem!
-        if not hasattr(self, "field"):
+        # We can only validate the expression if the modifier has been
+        # associated with a field and its form (we need to get the initial
+        # values from the form to test the expression).
+        try:
+            field_values = self.field.form.initial_values.dict()
+        except ObjectDoesNotExist:
             return
 
         # Validate that the expression is valid for the form. This is
         # accomplished by building a dict of initial values for all fields on
-        # the form and running it through the expression evaluator. If any
+        # the form (along with the initial values for concrete attributes on
+        # the Record) and running it through the expression evaluator. If any
         # exceptions are raised, they are returned as validation errors.
-        field_values = {f.name: f.initial for f in self.field.form.fields.all()}
+        Record = self.flexible_forms.get_model(BaseRecord)
+        record_attributes = {
+            str(f.attname): f.value_from_object(Record())
+            for f in cast(List[models.Field], Record._meta.get_fields())
+            if hasattr(f, "attname") and hasattr(f, "value_from_object")
+        }
+
+        expression_context: Mapping[str, Any] = {
+            "meta": record_attributes,
+            **field_values,
+        }
 
         try:
-            evaluate_expression(self.expression, names=field_values)
+            evaluate_expression(self.expression, names=expression_context)
 
         # If the expression references a name that isn't a field on the form
         # (or a builtin), raise a validation error.
         except NameNotDefined as ex:
-            valid_fields = ", ".join(field_values.keys())
+            valid_fields = ", ".join(expression_context.keys())
             name = getattr(ex, "name", "")
             raise ValidationError(
                 {
@@ -1019,14 +1066,14 @@ class RecordManager(models.Manager):
             (
                 super()
                 .get_queryset()
-                .select_related("_form")
+                .select_related("form")
                 .prefetch_related(
-                    f"_form__fields",
-                    f"_form__fields__modifiers",
-                    f"_form__fieldsets",
-                    f"_form__fieldsets__items",
+                    f"form__fields",
+                    f"form__fields__modifiers",
+                    f"form__fieldsets",
+                    f"form__fieldsets__items",
                     Prefetch(
-                        "_attributes",
+                        "attributes",
                         queryset=RecordAttribute.objects.select_related("field"),
                     ),
                 )
@@ -1049,49 +1096,49 @@ class BaseRecord(FlexibleBaseModel):
     names.
     """
 
-    _form: "models.ForeignKey[BaseForm, BaseForm]" = FlexibleForeignKey(
+    form: "models.ForeignKey[BaseForm, BaseForm]" = FlexibleForeignKey(
         BaseForm, on_delete=models.CASCADE, related_name="records"
     )
-    _form_id: Optional[int]
-    _attributes: "models.BaseManager[BaseRecordAttribute]"
+    form_id: Optional[int]
+    attributes: "models.BaseManager[BaseRecordAttribute]"
 
     objects = RecordManager()
 
     # The _initialized flag is used to determine if the model instance has been
     # fully initialized by Django. We use the flag to determine if it's safe to
-    # start proxying __getattr__ and __setattr__ calls to our _attributes
-    # association (which breaks if the instanec is not fully initialized).
+    # start proxying __getattr__ and __setattr__ calls to our attributes
+    # association (which breaks if the instance is not fully initialized).
     _initialized: bool = False
 
     # The __setattr__ proxy stores attribute updates in the _staged_changes
     # dict until save() is called. This attempts to mirror the way vanilla
     # Django models work.
-    _staged_changes: MutableMapping[str, Any]
+    _staged_changes: Dict[str, Any]
 
     class Meta:
         abstract = True
 
     class FlexibleMeta:
-        _form_field_name = "_form"
-        _form_field_related_name = "records"
+        form_field_name = "form"
+        form_field_related_name = "records"
 
     def __str__(self) -> str:
         # If we can get the name of the record's form without making a database
         # call, do it so we can make the __str__ more descriptive. Otherise,
         # just call it a "Record".
         record_type = (
-            self._form.label
-            if self.__dict__.get("_form")
-            or self._meta.get_field("_form").get_cached_value(self)
+            self.form.label
+            if self.__dict__.get("form")
+            or self._meta.get_field("form").get_cached_value(self)
             else "Record"
         )
 
         return f"New {record_type}" if not self.pk else f"{record_type} {self.pk}"
 
     def __init__(
-        self, *args: Any, _form: Optional[BaseForm] = None, **kwargs: Any
+        self, *args: Any, form: Optional[BaseForm] = None, **kwargs: Any
     ) -> None:
-        super().__init__(*args, _form=_form, **kwargs)
+        super().__init__(*args, form=form, **kwargs)
         self._staged_changes = {}
         self._initialized = True
 
@@ -1108,7 +1155,7 @@ class BaseRecord(FlexibleBaseModel):
             BaseRecordForm: A configured Django form for the record.
         """
         kwargs = {"instance": self, **kwargs}
-        return self._form.as_django_form(*args, **kwargs)
+        return self.form.as_django_form(*args, **kwargs)
 
     @cached_property
     def _fields(self) -> Mapping[str, BaseField]:
@@ -1118,9 +1165,9 @@ class BaseRecord(FlexibleBaseModel):
             Mapping[str, BaseField]: A mapping of Field instances by their
                 names.
         """
-        if not hasattr(self, "_form"):
+        if not hasattr(self, "form"):
             return {}
-        return {f.name: f for f in self._form.fields.all()}
+        return {f.name: f for f in self.form.fields.all()}
 
     @cached_property
     def _data(self) -> Mapping[str, Any]:
@@ -1132,11 +1179,7 @@ class BaseRecord(FlexibleBaseModel):
         initial_values = {name: field.initial for name, field in self._fields.items()}
         attribute_values = cast(
             Dict[str, Any],
-            (
-                {a.field.name: a.value for a in self._attributes.all()}
-                if self.pk
-                else {}
-            ),
+            ({a.field.name: a.value for a in self.attributes.all()} if self.pk else {}),
         )
         staged_attributes = {k: v for k, v in self._staged_changes.items()}
 
@@ -1230,7 +1273,7 @@ class BaseRecord(FlexibleBaseModel):
 
         # Upsert the record attributes.
         RecordAttribute = cast(Any, self._flexible_model_for(BaseRecordAttribute))
-        attribute_map = {a.field.name: a for a in self._attributes.all()}
+        attribute_map = {a.field.name: a for a in self.attributes.all()}
 
         value_fields: Set[str] = set()
         update: List[BaseRecordAttribute] = []
@@ -1278,7 +1321,7 @@ class BaseRecordAttribute(FlexibleBaseModel):
     record: "models.ForeignKey[BaseRecord, BaseRecord]" = FlexibleForeignKey(
         BaseRecord,
         on_delete=models.CASCADE,
-        related_name="_attributes",
+        related_name="attributes",
     )
     record_id: Optional[int]
 
@@ -1296,7 +1339,7 @@ class BaseRecordAttribute(FlexibleBaseModel):
         field_field_name = "field"
         field_field_related_name = "attributes"
         record_field_name = "record"
-        record_field_related_name = "_attributes"
+        record_field_related_name = "attributes"
 
     def __str__(self) -> str:
         return f"RecordAttribute {self.pk} (record_id={self.record_id}, field_id={self.field_id})"
@@ -1368,10 +1411,10 @@ class BaseRecordAttribute(FlexibleBaseModel):
 # storage by creating a column of an appropriate datatype for each supported
 # field.
 #
-for field_type, field in sorted(FIELD_TYPES.items(), key=lambda f: f[0]):
+for field_type_name, field_type in sorted(FIELD_TYPES.items(), key=lambda f: f[0]):
     BaseRecordAttribute.add_to_class(  # type: ignore
-        BaseRecordAttribute.get_value_field_name(field_type),
-        field.as_model_field(blank=True, null=True, default=None),
+        BaseRecordAttribute.get_value_field_name(field_type_name),
+        field_type.as_model_field(blank=True, null=True, default=None),
     )
 
 
@@ -1412,8 +1455,8 @@ class FlexibleForms:
         grade = models.PositiveIntegerField()
 
         class FlexibleMeta:
-            _form_field_name = "quiz"
-            _form_field_related_name = "submissions"
+            form_field_name = "quiz"
+            form_field_related_name = "submissions"
     ```
 
     ```python
@@ -1472,18 +1515,14 @@ class FlexibleForms:
         Args:
             model_class: The concrete implementation of a flexible base model.
         """
-        base_model = self._get_flexible_base_model(model_class)
-        base_opts = base_model._meta
-
-        self.models[base_model] = model_class
-        self.models[f"{base_opts.app_label}.{base_opts.model_name}"] = model_class
+        self.models[self._get_flexible_base_model(model_class)] = model_class
 
     def get_model(
         self,
-        base_model: Union[str, Type["FlexibleBaseModel"]],
+        base_model: Type[M],
         *,
         create: bool = False,
-    ) -> Type["FlexibleBaseModel"]:
+    ) -> Type[M]:
         """Returns the concrete model for the given flexible base model.
 
         Args:
@@ -1503,7 +1542,6 @@ class FlexibleForms:
         # If no model is registered for the requested base_model yet, and
         # create is True, create and register a new model.
         if base_model not in self.models and create:
-            base_model = cast(Type["FlexibleBaseModel"], base_model)
             self.register_model(
                 type(
                     base_model.__name__.replace("Base", self.model_prefix),
@@ -1513,7 +1551,7 @@ class FlexibleForms:
             )
 
         try:
-            return self.models[base_model]
+            return cast(Type[M], self.models[base_model])
         except KeyError:
             raise LookupError(
                 f"No concrete model has been registered for {base_model}: {self.models}"
@@ -1623,7 +1661,7 @@ class FlexibleForms:
         """
         return self._generate_model(BaseRecordAttribute)
 
-    def _generate_model(self, base_model: Type[T]) -> Type[T]:
+    def _generate_model(self, base_model: Type[M]) -> Type[M]:
         """Generate a model class from the given base_model.
 
         Args:
