@@ -2,13 +2,14 @@
 
 """Common utilities."""
 
+import json
 from typing import (
     TYPE_CHECKING,
     Any,
     List,
     Mapping,
     Optional,
-    Set,
+    Sequence,
     Tuple,
     Type,
     TypeVar,
@@ -39,22 +40,6 @@ if TYPE_CHECKING:  # pragma: no cover
 T = TypeVar("T", bound=Type)
 
 
-def all_subclasses(cls: T) -> Set[T]:
-    """Return a set of subclasses for the given class.
-
-    Recurses through the class hierarchy to find all descendants.
-
-    Args:
-        cls: The class for which to find all subclasses.
-
-    Returns:
-        Set[T]: The set of all descendants of `cls`.
-    """
-    return set(cls.__subclasses__()).union(
-        [s for c in cls.__subclasses__() for s in all_subclasses(c)]
-    )
-
-
 def empty(value: Any) -> bool:
     """Return True if the given value is "empty".
 
@@ -150,3 +135,57 @@ def replace_element(
             element = replace_element(needle, replacement, element)
         elements = type(haystack)([*elements, element])
     return elements
+
+
+def stable_json(data: Union[dict, list]) -> str:
+    """Generate a stable string representation of the given dict or list.
+
+    Args:
+        data: The dict or list for which to produce a stable JSON
+            representation.
+
+    Returns:
+        str: A stable JSON string representation of the given data.
+    """
+    return json.dumps(
+        data, sort_keys=True, ensure_ascii=True, separators=(",", ":"), default=str
+    )
+
+
+from string import Formatter
+
+
+class LenientFormatter(Formatter):
+    """A more lenient version of the default string formatter.
+
+    If a given variable was not specified in the format() kwargs, its value will be an empty string.
+
+    For example:
+
+    >>> formatter = LenientFormatter()
+    >>> test_string = "This variable was not provided: {not_provided}, but this one was: {provided}."
+    >>> formatter.format(test_string, provided="a value")
+    >>> 'This variable was not provided: , but this one was: a value.'
+    """
+
+    def get_value(
+        self, key: Union[int, str], args: Sequence[Any], kwargs: Mapping[str, Any]
+    ) -> Any:
+        """Return the value for the given key from the args or kwargs.
+
+        If the key does not exist in the args or kwargs, an empty string is
+        returned.
+
+        Args:
+            key: The name of the string variable for which to resolve the
+                value.
+            args: Positional arguments given to format().
+            kwargs: keyword arguments given to format().
+
+        Returns:
+            Any: The value for key, if given, otherwise an empty string.
+        """
+        try:
+            return super().get_value(key, args, kwargs)
+        except (KeyError, IndexError):
+            return ""
